#!/usr/bin/env python
"""
Test suite for the model
"""
from rethinkORM import RethinkModel


"""
################################################
Test Fixtures, setup and tear down run before the first test, and after the
last test, respectfully. These are responsible for initializing the RethinkDB
connection, making a test database, and later destroying that test database.

gateModel is the test object that we'll be working with
"""
# Sample data to use as a comparison as we test the model
baseData = {
    "what": "DHD",
    "description": """Dial Home Device from the planet P3X-439, where an
    Ancient Repository of Knowledge was found, and interfaced with by Colonel
    Jack.""",
    "id": "P3X-439-DHD",
    "planet": "P3X-439",
    "episodes": 66
    }


newData = {
    "what": "Star Gate",
    "description": """Device used to form a wormhole to another gate and
    transport matter.""",
    "id": "StarGate-Earth",
    "planet": "Earth",
    "episodes": 2,
    }

classmethodData = {
    "what": "Lost City, Part 1",
    "description": """The SG1 team finds an abandoned ancient statue which
    contains an ancient knowledge head.""",
    "id": 66
    }

secondJoinData = {
    "what": "Something",
    "description": "A forgotten episode.",
    "id": 2
    }


class gateModel(RethinkModel):
    """
    Sample document object which represents the documents within the table
    `stargates`.
    """
    table = "stargates"


class episodeModel(RethinkModel):
    """Sample document object to represent episode documents within the table
    `episodes`"""
    table = "episodes"



#!/usr/bin/env python
from rethinkModel import RethinkModel
from rethinkCollection import RethinkCollection
import tests

__version__ = '0.2.11'
VERSION = tuple(map(int, __version__.split('.')))

__all__ = ["RethinkModel",
           "RethinkCollection",
           "tests"]



#!/usr/bin/env python
"""
Quick way to get groupings of RethinkModels objects matching the given criteria
"""
import rethinkdb as r


class RethinkCollection(object):
    """
    A way to fetch groupings of documents that meet a criteria and have them
    in an iterable storage object, with each document represented by
    `RethinkModel` objects
    """
    documents = []
    table = ""
    _model = None
    _query = None
    _filter = {}
    _join = None
    _joinedField = None

    def __init__(self, model, filter=None):
        """
        Instantiates a new collection, using the given models table, and
        wrapping all documents with the given model.

        Filter can be a dictionary or lambda, similar to the filters for the
        RethinkDB drivers filters.
        """
        self._model = model
        self._query = r.table(self._model.table)

        if filter:
            self._filter = filter
            self._query = self._query.filter(self._filter)

    def joinOn(self, model, onIndex):
        """
        Performs an eqJoin on with the given model. The resulting join will be
        accessible through the models name.
        """
        return self._joinOnAsPriv(model, onIndex, model.__name__)

    def joinOnAs(self, model, onIndex, whatAs):
        """
        Like `joinOn` but allows setting the joined results name to access it
        from.

        Performs an eqJoin on with the given model. The resulting join will be
        accessible through the given name.
        """
        return self._joinOnAsPriv(model, onIndex, whatAs)

    def _joinOnAsPriv(self, model, onIndex, whatAs):
        """
        Private method for handling joins.
        """
        if self._join:
            raise Exception("Already joined with a table!")

        self._join = model
        self._joinedField = whatAs
        table = model.table
        self._query = self._query.eq_join(onIndex, r.table(table))
        return self

    def orderBy(self, field, direct="desc"):
        """
        Allows for the results to be ordered by a specific field. If given,
        direction can be set with passing an additional argument in the form
        of "asc" or "desc"
        """
        if direct == "desc":
            self._query = self._query.order_by(r.desc(field))
        else:
            self._query = self._query.order_by(r.asc(field))

        return self

    def __iter__(self):
        for doc in self._documents:
            yield doc

    # Pagination helpers...
    # These are questionable, on if I'll put them in or not.
    #def paginate(self, start,finish):
        #pass

    #@property
    #def currentPage(self):
        #pass

    #@property
    #def perpage(self):
        #pass

    #@property
    #def hasnextpage(self):
        #pass

    #@property
    #def pages(self):
        #pass
    # Okay, enough pagination

    def fetch(self):
        """
        Fetches the query and then tries to wrap the data in the model, joining
        as needed, if applicable.
        """
        returnResults = []

        results = self._query.run()
        for result in results:
            if self._join:
                # Because we can tell the models to ignore certian fields,
                # through the protectedItems blacklist, we can nest models by
                # name and have each one act normal and not accidentally store
                # extra data from other models
                item = self._model.fromRawEntry(**result["left"])
                joined = self._join.fromRawEntry(**result["right"])
                item.protectedItems = self._joinedField
                item[self._joinedField] = joined

            else:
                item = self._model.fromRawEntry(**result)

            returnResults.append(item)

        self._documents = returnResults
        return self._documents



#!/usr/bin/env python
"""
ORM style interface for working with RethinkDB and having a native wrapper and
some helper functions for working with the wrapper. This module contains
the base model which should be inherited.
"""
import rethinkdb as r


class RethinkModel(object):
    """
    Emulates a python object for the  data which is returned from rethinkdb and
    the official Python client driver. Raw data from the database is stored in
    _data to keep the objects namespace clean. For more information look at how
    _get() and _set() function in order to keep the namespace cleaner but still
    provide easy access to data.

    This object has a __repr__ method which can be used with print or logging
    statements. It will give the id and a representation of the internal _data
    dict for debugging purposes.
    """
    _protectedItems = []
    _conn = None
    _join = None
    _joinedField = ""

    table = ""  #: The table which this document object will be stored in
    primaryKey = "id"  #: The current primary key of the table

    durability = "soft"
    """Can either be Hard or Soft, and is passed to RethinkDB"""

    non_atomic = False
    """Determins if the transaction can be non atomic or not"""

    upsert = True
    """Will either update, or create a new object if true and a primary key is
    given."""

    def __init__(self, id=False, **kwargs):
        """
        Initializes the main object, if `id` is in kwargs, then we assume
        this is already in the database, and will try to pull its data, if not,
        then we assume this is a new entry that will be inserted.

        (Optional, only if not using .repl()) `conn` or `connection` can also
        be passed, which will be used in all the .run() clauses.
        """

        protectedItems = dir(self)
        protectedItems.append(self._protectedItems)
        self._protectedItems = protectedItems
        """
        List of strings to not store in the database; automatically set to
        the built in properties of this object to prevent any accidental stuff
        """

        # Is this a new object, or already in the database? (set later)
        self._new = True
        self._data = {}  # STORE ALL THE DATA!!

        # If we're given a connection, we'll use it, if not, we'll assume
        # .repl() was called on r.connect()
        self._conn = kwargs.pop("conn", kwargs.pop("connection", None))

        key = kwargs.get(self.primaryKey, id)

        if key is None or key == "" and len(kwargs) == 0:
            raise Exception("""Cannot have an empty or type None key""")

        elif key and len(kwargs) > 0:
            # Assume we have data from a collection, just go with it and set
            # our data.
            #self._makeNew(kwargs)
            raise Exception("""Cannot supply primary key and additional \
arguments while searching for Documents.""")

        if key and not self._grabData(key):
            raise Exception("""Could not find key in database""")

        self._makeNew(kwargs)
        if key:
            self._data[self.primaryKey] = key

        # Hook to run any inherited class code, if needed
        self.finishInit()

    def _makeNew(self, kwargs):
        # We assume this is a new object, and that we'll insert it
        for key in kwargs:
            if key not in object.__getattribute__(self, "_protectedItems") \
                   or key[0] != "_":
                self._data[key] = kwargs[key]

    def _grabData(self, key):
        """
        Tries to find the existing document in the database, if it is found,
        then the objects _data is set to that document, and this returns
        `True`, otherwise this will return `False`

        :param key: The primary key of the object we're looking for
        :type key: Str

        :return: True if a document was found, otherwise False
        :rtype: Boolean
        """
        rawCursor = r.table(self.table).get(key).run(self._conn)
        if rawCursor:
            self._data = rawCursor
            self._new = False
            return True
        else:
            return False

    def finishInit(self):
        """
        A hook called at the end of the main `__init__` to allow for
        custom inherited classes to customize their init process without having
        to redo all of the existing int.
        This should accept nothing besides `self` and nothing should be
        returned.
        """
        pass

    def _get(self, item):
        """
        Helper function to keep the __getattr__ and __getitem__ calls
        KISSish
        """
        if item not in object.__getattribute__(self, "_protectedItems") \
                and item[0] != "_":
            data = object.__getattribute__(self, "_data")
            if item in data:
                return data[item]
        return object.__getattribute__(self, item)

    def _set(self, item, value):
        """
        Helper function to keep the __setattr__ and __setitem__ calls
        KISSish

        Will only set the objects _data if the given items name is not prefixed
        with _ or if the item exists in the protected items List.
        """
        if item not in object.__getattribute__(self, "_protectedItems") \
                and item[0] != "_":
            keys = object.__getattribute__(self, "_data")
            if not hasattr(value, '__call__'):
                keys[item] = value
                return value
            if hasattr(value, '__call__') and item in keys:
                raise Exception("""Cannot set model data to a function, same \
name exists in data""")
        return object.__setattr__(self, item, value)

    def __getattr__(self, item):
        return self._get(item)

    def __getitem__(self, item):
        return self._get(item)

    def __setattr__(self, item, value):
        return self._set(item, value)

    def __setitem__(self, item, value):
        return self._set(item, value)

    def __delitem__(self, item):
        """
        Deletes the given item from the objects _data dict, or if from the
        objects namespace, if it does not exist in _data.
        """
        keys = object.__getattribute__(self, "_data")
        if item in keys:
            del(keys[item])
        else:
            object.__delitem__(self, item)

    def __contains__(self, item):
        """
        Allows for the use of syntax similar to::

            if "blah" in model:

        This only works with the internal _data, and does not include other
        properties in the objects namepsace.
        """
        keys = object.__getattribute__(self, "_data")
        if item in keys:
            return True
        return False

    @classmethod
    def fromRawEntry(cls, **kwargs):
        """
        Helper function to allow wrapping existing data/entries, such as
        those returned by collections.
        """
        id = kwargs["id"]

        kwargs.pop("id")

        what = cls(**kwargs)
        what._new = False
        what.id = id

        return what

    @classmethod
    def new(cls, **kwargs):
        """
        Creates a new instance, filling out the models data with the keyword
        arguments passed, so long as those keywords are not in the protected
        items array.
        """
        return cls(**kwargs)

    @classmethod
    def create(cls, id=None, **kwargs):
        """
        Similar to new() however this calls save() on the object before
        returning it.
        """
        what = cls(**kwargs)
        if id:
            setattr(what, cls.primaryKey, id)
        what.save()
        return what

    @classmethod
    def find(cls, id):
        """
        Loads an existing entry if one can be found, otherwise an exception is
        raised.

        :param id: The id of the given entry
        :type id: Str

        :return: `cls` instance of the given `id` entry
        """
        return cls(id)

    def save(self):
        """
        If an id exists in the database, we assume we'll update it, and if not
        then we'll insert it. This could be a problem with creating your own
        id's on new objects, however luckily, we keep track of if this is a new
        object through a private _new variable, and use that to determine if we
        insert or update.
        """
        if not self._new:
            data = self._data.copy()
            ID = data.pop(self.primaryKey)
            reply = r.table(self.table).get(ID) \
                .update(data,
                        durability=self.durability,
                        non_atomic=self.non_atomic) \
                .run(self._conn)

        else:
            reply = r.table(self.table) \
                .insert(self._data,
                        durability=self.durability,
                        upsert=self.upsert) \
                .run(self._conn)
            self._new = False

        if "generated_keys" in reply and reply["generated_keys"]:
            self._data[self.primaryKey] = reply["generated_keys"][0]

        if "errors" in reply and reply["errors"] > 0:
            raise Exception("Could not insert entry: %s"
                            % reply["first_error"])

        return True

    def delete(self):
        """
        Deletes the current instance. This assumes that we know what we're
        doing, and have a primary key in our data already. If this is a new
        instance, then we'll let the user know with an Exception
        """
        if self._new:
            raise Exception("This is a new object, %s not in data, \
indicating this entry isn't stored." % self.primaryKey)

        r.table(self.table).get(self._data[self.primaryKey]) \
            .delete(durability=self.durability).run(self._conn)
        return True

    def __repr__(self):
        """
        Allows for the representation of the object, for debugging purposes
        """
        return "< %s at %s with data: %s >" % (self.__class__.__name__,
                                               id(self),
                                               self._data)

    @property
    def protectedItems(self):
        """
        Provides a cleaner interface to dynamically add items to the models
        list of protected functions to not store in the database
        """
        return self._protectedItems

    @protectedItems.setter
    def protectedItems(self, value):
        if type(value) is list:
            self._protectedItems.extend(value)
        else:
            assert type(value) is str
            self._protectedItems.append(value)
        return self._protectedItems



#!/usr/bin/env python
"""
Test suite for the model
"""
from rethinkORM.tests.fixtures import *
from rethinkORM.rethinkCollection import RethinkCollection


class baseCollection_test(object):
    def normalCollection_test(self):
        one = gateModel.create(**baseData)
        two = gateModel.create(**newData)
        parts = [two, one]

        collection = RethinkCollection(gateModel)

        results = collection.fetch()
        assert len(results) == 2
        for model in range(len(results)):
            bit = parts[model]
            model = results[model]

            assert model.id == bit.id
            assert model.what == bit.what

        one.delete()
        two.delete()


class baseCollection_join_test(object):
    def joinCollection_test(self):
        one = gateModel.create(**baseData)
        two = gateModel.create(**newData)

        three = episodeModel.create(**classmethodData)
        four = episodeModel.create(**secondJoinData)

        collection = RethinkCollection(gateModel)
        collection.joinOn(episodeModel, "episodes")

        results = collection.fetch()
        assert len(results) == 2
        for result in results:
            assert result.episodeModel
            assert result.episodeModel.id == result.episodes

        one.delete()
        two.delete()
        three.delete()
        four.delete()


class baseCollection_joinAs_test(object):
    def joinCollection_test(self):
        one = gateModel.create(**baseData)
        two = gateModel.create(**newData)

        three = episodeModel.create(**classmethodData)
        four = episodeModel.create(**secondJoinData)

        collection = RethinkCollection(gateModel)
        collection.joinOnAs(episodeModel, "episodes", "epi")

        results = collection.fetch()
        assert len(results) == 2
        for result in results:
            assert result.epi
            assert result.epi.id == result.episodes

        one.delete()
        two.delete()
        three.delete()
        four.delete()


class baseCollection_orderBy_test(object):
    def orderBy_test(self):
        one = gateModel.create(**baseData)
        two = gateModel.create(**newData)

        three = episodeModel.create(**classmethodData)
        four = episodeModel.create(**secondJoinData)

        collection = RethinkCollection(gateModel)
        collection.orderBy('episodes', 'asc')
        results = collection.fetch()
        assert results[0].episodes < results[1].episodes

        collection = RethinkCollection(gateModel)
        collection.orderBy('episodes')
        results = collection.fetch()
        assert results[0].episodes > results[1].episodes

        one.delete()
        two.delete()
        three.delete()
        four.delete()


class baseCollection_filtering_test(object):
    """
    TODO: tests for:
      * Filtering
    """
    pass



import rethinkdb as r


def setup():
    """
    Sets up the connection to RethinkDB which we'll use in the rest of the
    tests, and puts it into .repl() mode so we don't have to pass the model
    object a connection. After that, we create a new database called `model`
    and within that a table called `stargate` and sets the database to use
    `model`.
    """
    conn = r.connect('localhost', 28015)
    r.db_create("model").run(conn)
    conn.use("model")
    conn.repl()
    r.table_create("stargates").run()
    r.table_create("episodes").run()

def teardown():
    """
    Drops the whole `model` database, since it's no longer needed now that
    the tests are done.
    """
    r.db_drop("model").run()



from distutils.core import setup
import rethinkORM

setup(
    name='RethinkORM',
    version=rethinkORM.__version__,
    author='Joshua P Ashby',
    author_email='joshuaashby@joshashby.com',
    packages=['rethinkORM', 'rethinkORM.tests'],
    url='https://github.com/JoshAshby/pyRethinkORM',
    license='GPL v3 (See LICENSE.txt for more info)',
    description='Useful little ORM style wrapper for working with RethinkDB',
    long_description=open('README.rst').read(),
    install_requires=[
        "rethinkdb >= 1.7.0",
        "nose >= 1.3.0"
    ],
)



#!/usr/bin/env python
"""
Test suite for the model
"""
import nose.tools as nst
from rethinkORM.tests.fixtures import *


class base(object):
    """
    Base test object to help automate some of the repetitive work of reloading
    a document to ensure the model matches the test data. Also takes care of
    deleting the document if `cleanupAfter` is `True`
    """
    cleanupAfter = False
    """Should the document created by this test be deleted when over?"""
    loadCheck = True
    """
    Should the document be reloaded and have all it's data checked against?
    """
    whatToLoad = []
    """
    If loadCheck is true, fill this out with strings of the data keys
    to check the model against.
    """

    model = None
    """The model being used for this test"""
    data = None
    """The data being used for this test. Please at least include an ID"""

    def action_test(self):
        self.action()

    # prefixed with b_ to run this right after action because things run alpha
    # order it seems
    def b_load_test(self):
        if self.loadCheck:
            self.load()

    def cleanup_test(self):
        if self.cleanupAfter:
            self.cleanup()

    def action(self):
        """
        Override this with your own function to do whatever you want for the
        test
        """
        pass

    def load(self):
        """
        Override this to do a custom load check. This should find the key you
        created or modified in `action()` and check it's values to ensure
        everything was set correctly. By default this loads the model with the
        test objects `data["id"]` and uses `whatToLoad` to run checks against
        the data and the model.
        """
        item = self.model(self.data["id"])
        assert item.id == self.data["id"]
        for bit in self.whatToLoad:
            assert getattr(item, bit) == self.data[bit]
        del item

    def cleanup(self):
        """
        Override this to set a custom cleanup process. By default this takes
        the key that was generated in `action()` and calls the models
        `.delete()` function.
        """
        item = self.model(self.data["id"])
        item.delete()
        del item


class insert_test(base):
    """
    Tests the basic ability to make a new model instance, and save it to the
    Database
    """
    whatToLoad = ["what", "description", "planet"]
    data = {"what": "DHD",
            "description": "Dial Home Device",
            "planet": baseData["planet"],
            "id": baseData["id"]}
    model = gateModel

    cleanupAfter = True

    def action(self):
        """
        Creates a new object, and inserts it, using `.save()`
        """
        dhdProp = gateModel(what=self.data["what"],
                            description=self.data["description"],
                            planet=self.data["planet"])
        dhdProp.id = self.data["id"]
        assert dhdProp.save()
        del dhdProp


class modify_test(base):
    """Tests the ability to load, modify and save a model correctly"""
    whatToLoad = ["what", "description", "planet", "episodes"]
    data = baseData
    model = gateModel

    cleanupAfter = True

    def action(self):
        """
        Next, we get the object again, and this time,
        we modify it, and save it.
        """
        gateModel.create(**{"what": "DHD",
            "description": "Dial Home Device",
            "planet": baseData["planet"],
            "id": self.data["id"]})

        dhdProp = gateModel(self.data["id"])
        dhdProp.what = self.data["what"]
        dhdProp.description = self.data["description"]
        dhdProp.episodes = self.data["episodes"]
        assert dhdProp.save()
        del dhdProp

        @nst.raises(Exception)
        def d_load_delete_test(self):
            """
            And make sure that if we try to get that object after it's been
            deleted, that we get a new object rather than the existing
            one we deleted.
            """
            dhdProp = self.model(self.data["id"])


@nst.raises(Exception)
def insertBadId_test():
    """
    Here we test to make sure that if we give a primary key of type `None`
    that we are raising an exception, if we don't get an exception then
    something is wrong since the primary key shouldn't be allowed to be `None`
    """
    oldProp = gateModel(id=None, what="Something?")
    assert oldProp.save()
    del oldProp


@nst.raises(Exception)
def insertIdAndData_test():
    """
    Make sure that the model raises an Exception when a key and data are
    provided
    """
    prop = gateModel(id="3", what="duh")


"""
Now onto the classmethods and helper functions to ensure things are good to go
"""


class new_classmethod_test(base):
    """
    Tests the new() classmethod of the model
    """
    model = gateModel
    data = classmethodData
    whatToCheck = ["what", "description"]

    cleanupAfter = True

    def action(self):
        prop = gateModel.new(what=self.data["what"],
                             description=self.data["description"])
        prop.id = self.data["id"]
        assert prop.what == self.data["what"]
        assert prop.description == self.data["description"]
        assert prop.id == self.data["id"]
        assert prop.save()


class create_classmethod_test(base):
    """
    Tests the create() classmethod of the model

    Same as the new() classmethod test however we don't have to explicitly
    tell the model to save
    """
    model = gateModel
    data = classmethodData
    whatToCheck = ["what", "description"]
    cleanupAfter = True

    def action(self):
        prop = gateModel.create(what=self.data["what"],
                                description=self.data["description"],
                                id=self.data["id"])
        assert prop.what == self.data["what"]
        assert prop.description == self.data["description"]
        assert prop.id == self.data["id"]


class find_classmethod_test(base):
    """
    Tests the find() classmethod of the model
    """
    model = gateModel
    data = classmethodData
    whatToCheck = ["what", "description"]
    loadCheck = False
    cleanupAfter = True

    def action(self):
        oldProp = gateModel(what=self.data["what"],
                            description=self.data["description"])
        oldProp.id = self.data["id"]
        oldProp.save()

        prop = gateModel.find(self.data["id"])
        assert prop.what == self.data["what"]
        assert prop.description == self.data["description"]
        assert prop.id == self.data["id"]
